# Copyright (c) SI-Analytics. All rights reserved.
from .base import (BaseSpace, Lograndint, Loguniform, Qlograndint, Qloguniform,
                   Qrandint, Qrandn, Quniform, Randint, Randn, Uniform)
from .builder import SPACES, build_space
from .choice import Choice
from .grid_search import GridSearch
from .sample_from import SampleFrom

__all__ = [
    'BaseSpace', 'Uniform', 'Quniform', 'Loguniform', 'Qloguniform', 'Randn',
    'Qrandn', 'Randint', 'Qrandint', 'Lograndint', 'Qlograndint', 'SPACES',
    'build_space', 'Choice', 'GridSearch', 'SampleFrom'
]



from setuptools import find_packages, setup

version_file = 'siatune/version.py'


def parse_requirements(fname: str = 'requirements.txt',
                       with_version: bool = True):
    """Parse the package dependencies listed in a requirements file but strips
    specific versioning information.

    Args:
        fname (str): path to requirements file
        with_version (bool, default=False): if True include version specs
    Returns:
        List[str]: list of requirements items
    CommandLine:
        python -c "import setup; print(setup.parse_requirements())"
    """
    import re
    import sys
    from os.path import exists
    require_fpath = fname

    def parse_line(line: str):
        """Parse information from a line in a requirements text file.

        Args:
            line (str): a line in a requirements text file
        Yields:
            Dict: a dictionary of parsed information
        """
        if line.startswith('-r '):
            # Allow specifying requirements in other files
            target = line.split(' ')[1]
            for info in parse_require_file(target):
                yield info
        else:
            info = {'line': line}
            if line.startswith('-e '):
                info['package'] = line.split('#egg=')[1]
            elif '@git+' in line:
                info['package'] = line
            else:
                # Remove versioning from the package
                pat = '(' + '|'.join(['>=', '==', '>']) + ')'
                parts = re.split(pat, line, maxsplit=1)
                parts = [p.strip() for p in parts]

                info['package'] = parts[0]
                if len(parts) > 1:
                    op, rest = parts[1:]
                    if ';' in rest:
                        # Handle platform specific dependencies
                        # http://setuptools.readthedocs.io/en/latest/setuptools.html#declaring-platform-specific-dependencies
                        version, platform_deps = map(str.strip,
                                                     rest.split(';'))
                        info['platform_deps'] = platform_deps
                    else:
                        version = rest  # NOQA
                    info['version'] = (op, version)
            yield info

    def parse_require_file(fpath: str):
        """Parse the package dependencies in a requirements file.

        Args:
            fpath (str): path to requirements file

        Yields:
            Dict: a dictionary of parsed information
        """
        with open(fpath, 'r') as f:
            for line in f.readlines():
                line = line.strip()
                if line and not line.startswith(('#', '-i')):
                    for info in parse_line(line):
                        yield info

    def gen_packages_items() -> str:
        """Generated package items.

        Yields:
            str: package items
        """

        if exists(require_fpath):
            for info in parse_require_file(require_fpath):
                parts = [info['package']]
                if with_version and 'version' in info:
                    parts.extend(info['version'])
                if not sys.version.startswith('3.4'):
                    # apparently package_deps are broken in 3.4
                    platform_deps = info.get('platform_deps')
                    if platform_deps is not None:
                        parts.append(';' + platform_deps)
                item = ''.join(parts)
                yield item

    packages = list(gen_packages_items())
    return packages


def get_version() -> str:
    """Get the version string from the package.

    Returns:
        str: The version string.
    """
    with open(version_file, 'r') as f:
        exec(compile(f.read(), version_file, 'exec'))
    return locals()['__version__']


if __name__ == '__main__':
    setup(
        name='siatune',
        version=get_version(),
        author='SIA',
        packages=find_packages(exclude=('configs', 'tools')),
        install_requires=parse_requirements('requirements/runtime.txt'),
        extras_require={
            'optional': parse_requirements('requirements/optional.txt'),
        },
    )



# Copyright (c) SI-Analytics. All rights reserved.
from .builder import TRIAL_SCHEDULERS, build_scheduler
from .pbt import PopulationBasedTraining

__all__ = ['TRIAL_SCHEDULERS', 'build_scheduler', 'PopulationBasedTraining']



# Copyright (c) SI-Analytics. All rights reserved.
import os


def set_env_vars(rank: int,
                 world_size: int,
                 addr: str = '127.0.0.1',
                 port: int = 29500):
    """Sets environment variables for distributed training.

    Args:
        rank (int):
            The rank of the current process.
        world_size (int):
            The total number of processes
            in the distributed training.
        addr (str, optional):
            The address of the master process.
            Defaults to '127.0.0.1'.
        port (int, optional):
            The port number used by the master process.
            Defaults to 29500.
    """
    os.environ['MASTER_ADDR'] = addr
    os.environ['MASTER_PORT'] = str(port)
    os.environ['RANK'] = str(rank)
    os.environ['WORLD_SIZE'] = str(world_size)



# Copyright (c) SI-Analytics. All rights reserved.
import inspect

from mmengine.config import Config
from mmengine.registry import Registry
from ray import tune
from ray.tune.schedulers import TrialScheduler

TRIAL_SCHEDULERS = Registry('trial scheduler')

# Dynamically import scheduler
# Refer to https://github.com/ray-project/ray/blob/master/python/ray/tune/schedulers/__init__.py  # noqa
for cls in set(tune.schedulers.SCHEDULER_IMPORT.values()):
    if not inspect.isclass(cls):
        continue
    TRIAL_SCHEDULERS.register_module(module=cls)


def build_scheduler(cfg: Config) -> TrialScheduler:
    """Build the scheduler from configs.

    Args:
        cfg (Config): The configs.
    Returns:
        tune.schedulers.TrialScheduler: The scheduler.
    """

    return TRIAL_SCHEDULERS.build(cfg)



# Copyright (c) SI-Analytics. All rights reserved.
from .builder import CALLBACKS, build_callback
from .mlflow import MLflowLoggerCallback

__all__ = ['CALLBACKS', 'build_callback', 'MLflowLoggerCallback']



# Copyright (c) SI-Analytics. All rights reserved.
from mmengine.config.config import DELETE_KEY, Config, ConfigDict

from .base import BaseRewriter
from .builder import REWRITERS


@REWRITERS.register_module()
class MergeConfig(BaseRewriter):
    """Merge the two configs."""

    def __init__(self, src_key: str, dst_key: str, key: str):
        """Initialize the MergeConfig class.

        Args:
            src_key (str): The key of the configs.
            dst_key (str): The key of the configs.
            key (str):
                The context key where the merged config will be stored.
        """
        self.src_key = src_key
        self.dst_key = dst_key
        self.key = key

    @staticmethod
    def merge_dict(src: dict, dst: dict, allow_list_keys: bool = False):
        """merge dict ``a`` into dict ``b`` (non-inplace).
        Values in ``a`` will overwrite ``b``. ``b`` is copied first to avoid
        in-place modifications.
        Args:
            a (dict): The source dict to be merged into ``b``.
            b (dict): The origin dict to be fetch keys from ``a``.
            allow_list_keys (bool): If True, int string keys (e.g. '0', '1')
              are allowed in source ``a`` and will replace the element of the
              corresponding index in b if b is a list. Default: False.
        Returns:
            dict: The modified dict of ``b`` using ``a``.
        Examples:
            # Normally merge a into b.
            >>> Config._merge_a_into_b(
            ...     dict(obj=dict(a=2)), dict(obj=dict(a=1)))
            {'obj': {'a': 2}}
            # Delete b first and merge a into b.
            >>> Config._merge_a_into_b(
            ...     dict(obj=dict(_delete_=True, a=2)), dict(obj=dict(a=1)))
            {'obj': {'a': 2}}
            # b is a list
            >>> Config._merge_a_into_b(
            ...     {'0': dict(a=2)}, [dict(a=1), dict(b=2)], True)
            [{'a': 2}, {'b': 2}]

        """
        dst = dst.copy()
        for k, v in src.items():
            if allow_list_keys and k.isdigit() and isinstance(dst, list):
                k = int(k)
                if len(dst) <= k:
                    raise KeyError(
                        f'Index {k} exceeds the length of list {dst}')
                # modified from the mmcv.config.Config._merge_a_into_b
                # this allows merging with primitives such as int, float
                dst[k] = MergeConfig.merge_dict(v, dst[k],
                                                allow_list_keys) if hasattr(
                                                    dst[k], 'copy') else v
            elif isinstance(v, dict):
                if k in dst and not v.pop(DELETE_KEY, False):
                    allowed_types = (dict, list) if allow_list_keys else dict
                    if not isinstance(dst[k], allowed_types):
                        raise TypeError(
                            f'{k}={v} in child config cannot inherit from '
                            f'base because {k} is a dict in the child config '
                            f'but is of type {type(dst[k])} in base config. '
                            f'You may set `{DELETE_KEY}=True` to ignore the '
                            f'base config.')
                    dst[k] = MergeConfig.merge_dict(v, dst[k], allow_list_keys)
                else:
                    dst[k] = ConfigDict(v)
            else:
                dst[k] = v
        return dst

    def __call__(self, context: dict, allow_list_keys: bool = True) -> dict:
        """Semantically merge and save the two configs.

        Args:
            context (dict): The context to be rewritten.
            allow_list_keys (bool): If True, int string keys (e.g. '0', '1')
                are allowed in source ``a`` and will replace the element of the
                corresponding index in b if b is a list. Default: True.

        Returns:
            dict: The context after rewriting.
        """
        src = context.pop(self.src_key)
        dst = context.pop(self.dst_key)
        unpacked_src = {}
        for full_key, v in src.items():
            d = unpacked_src
            key_list = full_key.split('.')
            for subkey in key_list[:-1]:
                d.setdefault(subkey, ConfigDict())
                d = d[subkey]
            subkey = key_list[-1]
            d[subkey] = v
        context[self.key] = Config(
            self.merge_dict(
                unpacked_src,
                dst.__getattribute__('_cfg_dict'),
                allow_list_keys=allow_list_keys),
            cfg_text=dst.text,
            filename=dst.filename)
        return context



# Copyright (c) SI-Analytics. All rights reserved.
import inspect

from mmengine.config import Config
from mmengine.registry import Registry
from ray import tune

STOPPERS = Registry('stopper')
for stopper in dir(tune.stopper):
    if not stopper.endswith('Stopper'):
        continue
    stopper_cls = getattr(tune.stopper, stopper)
    if not inspect.isclass(stopper_cls):
        continue
    STOPPERS._register_module(stopper_cls)


def build_stopper(cfg: Config) -> tune.stopper:
    """Build a stopper.

    Args:
        cfg (Config): The configuration of the stopper.

    Returns:
        tune.stopper: The instantiated stopper.
    """
    return STOPPERS.build(cfg)



# Copyright (c) SI-Analytics. All rights reserved.
from abc import ABCMeta, abstractmethod
from typing import Dict


class BaseRewriter(metaclass=ABCMeta):
    """Base class of rewriters."""

    @abstractmethod
    def __call__(self, context: Dict) -> Dict:
        """Rewrite the context.

        Args:
            context (Dict): The context to be rewritten.

        Returns:
            Dict: The context after rewriting.
        """
        return context
